
import sys
from math import inf
from collections import defaultdict
import random
import pickle


class Tagger:
    
    def __init__(self, arg):
        """ Im Training wird die maximale Suffixlänge als Argument übergeben
            und beim Taggen der Name der Parameterdatei
        """
        if type(arg) == str: # Taggen
            with open(arg, "rb") as file:
                self.max_suffix_length, self.tagset, self.weights = \
                                pickle.load(file)
        else:  # Training
            print("Training")
            self.max_suffix_length = arg
            self.weights = defaultdict(float)

            
    @staticmethod
    def read_data(filename):
        """ liest eine Datei ein, welche in jeder Zeile ein word
            und ein Tage enthält mit einer Leerzeile nach jedem Satz 
        """
        words, tags, data = [], [], []
        tagset = set()
        with open(filename) as file:
            for line in file:
                line = line.strip()
                if line == "":  # Satzende
                    # Grenzzeichen hinzufügen
                    data.append((words, tags))
                    words, tags = [], []
                else: 
                    word, tag = line.split("\t")
                    words.append(word)
                    tags.append(tag)
                    tagset.add(tag)
        tagset.add("<s>")
        return data, tagset

    
    def get_features(self, prevtag, tag, word):
        """ Berechnung des Merkmalsvektors als Liste von
            Merkmalsnamen, deren Wert jeweils 1 ist 
        """
        features = [f"TT-{prevtag}-{tag}",
                    f"TW-{tag}-{word}"]
        if word:
            case = "k" if word[0].islower() else \
                "g" if word[0].isupper() else \
                "0"
            for i in range(min(self.max_suffix_length+1, len(word))):
                suffix = word[-i:]
                features.append(f"TS-{tag}-{suffix}-{case}")
        return features
    
    
    def local_score(self, features):
        """ Multiplikation von Merkmals- und Gewichtsvektor """
        return sum(self.weights.get(f, 0.0) for f in features)
    
    
    def viterbi(self, words):
        """ Berechnung der besten Tagfolge """
        words = [""] + words + [""]
        vitscore = [dict() for _ in words]
        bestprev = [dict() for _ in words]
        vitscore[0]["<s>"] = 0
        for i in range(1, len(words)):
            for tag in self.tagset:
                for prevtag, prevscore in vitscore[i-1].items():
                    features = self.get_features(prevtag, tag, words[i])
                    score = prevscore + self.local_score(features)
                    if score > vitscore[i].get(tag, -inf):
                        vitscore[i][tag] = score
                        bestprev[i][tag] = prevtag
        tags = []
        tag = "<s>"
        for i in range(len(words)-1, 1, -1):
            tag = bestprev[i][tag]
            tags.append(tag)
        return tags[::-1]
    
    
    def update_weights(self, words, goldtags, tags):
        """ Anpassung des Gewichtsvektors """
        words = [""] + words + [""]  # Grenztokens hinzufügen
        tags = ["<s>"] + tags + ["<s>"]  # Grenztags hinzufügen
        goldtags = ["<s>"] + goldtags + ["<s>"]  # Grenztags hinzufügen
        for i in range(1, len(words)):
            if goldtags[i-1] != tags[i-1] or goldtags[i] != tags[i]:
                features = self.get_features(goldtags[i-1], goldtags[i], words[i])
                for f in features:
                    self.weights[f] += 1
                features = self.get_features(tags[i-1], tags[i], words[i])
                for f in features:
                    self.weights[f] -= 1

                    
    def train(self, filename, paramfile, num_iterations):
        """ Hauptprogramm für das Training """
        data, self.tagset = self.read_data(filename)
        for i in range(1, num_iterations+1):
            print(i,end="\r")
            words, goldtags = random.choice(data)
            tags = self.viterbi(words)
            self.update_weights(words, goldtags, tags)
            if i%100 == 0:
                print(*words[1:-1])
                print(*goldtags[1:-1])
                print(*tags[1:-1])
            if i%1000 == 0:
                print("saving parameters")
                with open(f"{paramfile}-{i}.pkl", "wb") as file:
                    params = [self.max_suffix_length,
                              self.tagset, dict(self.weights)] 
                    pickle.dump(params, file)

                    
if __name__ == "__main__":
    tagger = Tagger(5)
    tagger.train(sys.argv[1], sys.argv[2], 20000)
